@@ -20,11 +20,13 @@ import pathlib
 from typing import List, Optional, Set
 
 from kfp_tekton.compiler._k8s_helper import sanitize_k8s_name
-from kfp_tekton.compiler._op_to_template import _get_base_step, _add_mount_path, _prepend_steps
+from kfp_tekton.compiler._op_to_template import _get_base_step, _add_mount_path, _prepend_steps, _update_volumes
 from os import environ as env
 
 BIG_DATA_MIDPATH = "artifacts/$ORIG_PR_NAME"
 BIG_DATA_PATH_FORMAT = "/".join(["$(workspaces.$TASK_NAME.path)", BIG_DATA_MIDPATH, "$TASKRUN_NAME", "$TASK_PARAM_NAME"])
+ARTIFACT_OUTPUTLIST_ANNOTATION_KEY = 'artifact_outputs'
+OUTPUT_RESULT_PATH_SUFFIX = '-datap'
 
 
 def fix_big_data_passing(workflow: dict, loops_pipeline: dict, loop_name_prefix: str) -> dict:
@@ -323,6 +325,7 @@ def fix_big_data_passing(workflow: dict, loops_pipeline: dict, loop_name_prefix:
                 ),  # TODO: pipeline has no name, use pipelineRun name?
                 input_parameter['name']) in inputs_consumed_as_parameters
                 or input_parameter['name'].endswith("-trname")
+                or input_parameter['name'].endswith(OUTPUT_RESULT_PATH_SUFFIX)
         ]
 
     # Remove output parameters unless they're used downstream
@@ -331,11 +334,12 @@ def fix_big_data_passing(workflow: dict, loops_pipeline: dict, loop_name_prefix:
         spec['results'] = [
             output_parameter for output_parameter in spec.get('results', [])
         ]
+        spec['results'] = sorted(spec['results'], key=lambda k: k['name'])
         # tekton results doesn't support underscore
         renamed_results_in_pipeline_task = set()
         for task_result in spec['results']:
             task_result_old_name = task_result.get('name')
-            task_result_new_name = sanitize_k8s_name(task_result_old_name)
+            task_result_new_name = sanitize_k8s_name(task_result_old_name, allow_capital=True)
             if task_result_new_name != task_result_old_name:
                 task_result['name'] = task_result_new_name
                 renamed_results_in_pipeline_task.add(
@@ -361,6 +365,7 @@ def fix_big_data_passing(workflow: dict, loops_pipeline: dict, loop_name_prefix:
                 parameter_argument['name']) not in inputs_consumed_as_artifacts
                 or task['name'] in resource_template_names
                 or parameter_argument['name'].endswith("-trname")
+                or parameter_argument['name'].endswith(OUTPUT_RESULT_PATH_SUFFIX)
             ]
 
         # tekton results doesn't support underscore
@@ -373,7 +378,7 @@ def fix_big_data_passing(workflow: dict, loops_pipeline: dict, loop_name_prefix:
                 argument['value'] = '$(tasks.%s.%s.%s)' % (
                     argument_placeholder_parts[1],
                     argument_placeholder_parts[2],
-                    sanitize_k8s_name(argument_placeholder_parts[3]))
+                    sanitize_k8s_name(argument_placeholder_parts[3], allow_capital=True))
 
     workflow = jsonify_annotations(workflow)
     # Need to confirm:
@@ -446,6 +451,22 @@ def big_data_passing_pipeline(name: str, template: dict, inputs_tasks: set(),
                     "name": task.get('name'),
                     "workspace": pipeline_name
                 })
+        artifact_output_list = task.get('taskSpec', {}).get('metadata', {}).get('annotations', {}).get(
+            ARTIFACT_OUTPUTLIST_ANNOTATION_KEY, '')
+        if artifact_output_list:
+            tmp_list = set()
+            for output in json.loads(artifact_output_list):
+                tmp_list.add(sanitize_k8s_name(output, allow_capital=True))
+            for task_output in task.get('taskSpec', {}).get('results', []):
+                if task_output.get('name') in tmp_list:
+                    if not task.setdefault('workspaces', []):
+                        task['workspaces'].append({
+                          "name": task.get('name'),
+                            "workspace": pipeline_name
+                        })
+                        pipeline_workspaces.add(pipeline_name)
+                        break
+
     if pipeline_name in pipeline_workspaces:
         # Add workspaces to pipeline
         if not pipeline_spec.setdefault('workspaces', []):
@@ -488,25 +509,45 @@ def big_data_passing_tasks(prname: str, task: dict, pipelinerun_template: dict,
     task_spec = task.get('taskSpec', {})
     # Data passing for the task outputs
     appended_taskrun_name = False
+    appended_taskrun_path_step = None
+    artifact_output_list = task_spec.get('metadata', {}).get('annotations', {}).get(ARTIFACT_OUTPUTLIST_ANNOTATION_KEY, '')
+    if artifact_output_list:
+        temp_list = json.loads(artifact_output_list)
+        artifact_output_list = []
+        for output in temp_list:
+            artifact_output_list.append(sanitize_k8s_name(output, allow_capital=True))
     for task_output in task.get('taskSpec', {}).get('results', []):
-        if (task_name, task_output.get('name')) in outputs_tasks:
+        if (task_name, task_output.get('name')) in outputs_tasks or \
+            (artifact_output_list and task_output.get('name') in artifact_output_list):
             if not task.get('taskSpec', {}).setdefault('workspaces', []):
                 task.get('taskSpec', {})['workspaces'].append({"name": task_name})
             # Replace the args for the outputs in the task_spec
             # $(results.task_output.get('name').path)  -->
             # $(workspaces.task_name.path)/task_name-task_output.get('name')
             placeholder = '$(results.%s.path)' % (sanitize_k8s_name(
-                task_output.get('name')))
+                task_output.get('name'), allow_capital=True))
             workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
                 task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_output.get('name'))
-            # For child nodes to know the taskrun name, it has to pass to results via /tekton/results emptydir
-            if not appended_taskrun_name:
-                copy_taskrun_name_step = _get_base_step('output-taskrun-name')
-                copy_taskrun_name_step['script'] += 'echo -n "%s" > $(results.taskrun-name.path)\n' % ("$(context.taskRun.name)")
-                task['taskSpec']['results'].append({"name": "taskrun-name"})
-                task['taskSpec']['steps'].append(copy_taskrun_name_step)
-                _append_original_pr_name_env(task)
-                appended_taskrun_name = True
+            if env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
+            # For child nodes to know the taskrun output path, it has to pass to results via /tekton/results emptydir
+                if not appended_taskrun_path_step:
+                    appended_taskrun_path_step = _get_base_step('output-taskrun-path')
+                if len(appended_taskrun_path_step['command']) <= 2:
+                    appended_taskrun_path_step['command'].append('')
+                appended_taskrun_path_step['command'][-1] += 'echo -n "%s/%s/%s" > $(results.%s%s.path)\n' % \
+                                                    (BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_output.get('name'),
+                                                    task_output.get('name'), OUTPUT_RESULT_PATH_SUFFIX)
+                task['taskSpec']['results'].append({"name": "%s%s" % (task_output.get('name'), OUTPUT_RESULT_PATH_SUFFIX),
+                                                    "type": "string"})
+            else:
+                # For child nodes to know the taskrun name, it has to pass to results via /tekton/results emptydir
+                if not appended_taskrun_name:
+                    copy_taskrun_name_step = _get_base_step('output-taskrun-name')
+                    copy_taskrun_name_step['command'].append('echo -n "$(context.taskRun.name)" > "$(results.taskrun-name.path)"')
+                    task['taskSpec']['results'].append({"name": "taskrun-name", "type": "string"})
+                    task['taskSpec']['steps'].append(copy_taskrun_name_step)
+                    _append_original_pr_name_env(task)
+                    appended_taskrun_name = True
             task['taskSpec'] = replace_big_data_placeholder(
                 task.get("taskSpec", {}), placeholder, workspaces_parameter)
             artifact_items = pipelinerun_template['metadata']['annotations']['tekton.dev/artifact_items']
@@ -514,6 +555,9 @@ def big_data_passing_tasks(prname: str, task: dict, pipelinerun_template: dict,
                 artifact_items[task['name']], placeholder, workspaces_parameter)
             pipelinerun_template['metadata']['annotations']['tekton.dev/artifact_items'] = \
                 artifact_items
+    if appended_taskrun_path_step:
+        task['taskSpec']['steps'].append(appended_taskrun_path_step)
+        _append_original_pr_name_env(task)
 
     task_spec = task.get('taskSpec', {})
     task_params = task_spec.get('params', [])
@@ -541,34 +585,48 @@ def big_data_passing_tasks(prname: str, task: dict, pipelinerun_template: dict,
             # If the param name is constructed with task_name-param_name,
             # use the current task_name as the path prefix
 
-            def append_taskrun_params(task_name_append: str):
+            def append_taskrun_params(task_name_append: str, task_path_name: str):
                 taskrun_param_name = task_name_append + "-trname"
                 inserted_taskrun_param = False
-                for param in task['taskSpec'].get('params', []):
-                    if param.get('name', "") == taskrun_param_name:
-                        inserted_taskrun_param = True
-                        break
-                if not inserted_taskrun_param:
-                    task['taskSpec']['params'].append({"name": taskrun_param_name})
-                    task['params'].append({"name": taskrun_param_name, "value": "$(tasks.%s.results.taskrun-name)" % task_name_append})
+                param_name = {"name": taskrun_param_name}
+                param_nested_name = task_name_append + '-taskrun-name'
+                param_content = {"name": taskrun_param_name, "value": "$(tasks.%s.results.taskrun-name)" % task_name_append}
+                param_nested_content = {'name': task_name_append + '-taskrun-name',
+                                        'value': '$(tasks.%s.results.taskrun-name)' % task_name_append}
+                if env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
+                    param_path_name = '-'.join([task_name_append, task_path])
+                    param_name = {"name": param_path_name}
+                    param_nested_name = param_path_name
+                    param_content = {"name": param_path_name,
+                                     "value": "$(tasks.%s.results.%s)" % (task_name_append, task_path_name)}
+                    param_nested_content = param_content
+                else:
+                    for param in task['taskSpec'].get('params', []):
+                        if param.get('name', "") == taskrun_param_name:
+                            inserted_taskrun_param = True
+                            break
+                if (not inserted_taskrun_param) or env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
+                    task['taskSpec']['params'].append(param_name)
+                    task['params'].append(param_content)
                     parent_task_queue = [task['name']]
                     while parent_task_queue:
                         current_task = parent_task_queue.pop(0)
                         for loop_name, loop_spec in loops_pipeline.items():
-                            # print(loop_name, loop_spec)
                             if current_task in loop_spec.get('task_list', []):
                                 parent_task_queue.append(loop_name.replace(loop_name_prefix, ""))
                                 loop_param_names = [loop_param['name'] for loop_param in loops_pipeline[loop_name]['spec']['params']]
-                                if task_name_append + '-taskrun-name' in loop_param_names:
+                                if param_nested_name in loop_param_names:
                                     continue
-                                loops_pipeline[loop_name]['spec']['params'].append({'name': task_name_append + '-taskrun-name',
-                                'value': '$(tasks.%s.results.taskrun-name)' % task_name_append})
+                                loops_pipeline[loop_name]['spec']['params'].append(param_nested_content)
 
             if task_param_task_name:
                 workspaces_parameter = '$(workspaces.%s.path)/%s/$(params.%s-trname)/%s' % (
                     task_name, BIG_DATA_MIDPATH, task_param_task_name, task_param_param_name)
+                task_path = sanitize_k8s_name(task_param_param_name, allow_capital=True) + OUTPUT_RESULT_PATH_SUFFIX
+                if env.get('OUTPUT_BIG_DATA_PATH', 'false').lower() == 'true':
+                    workspaces_parameter = '$(workspaces.%s.path)/$(params.%s)' % (task_name, '-'.join([task_param_task_name, task_path]))
                 if task_param_task_name != task_name:
-                    append_taskrun_params(task_param_task_name)  # need to get taskrun name from parent path
+                    append_taskrun_params(task_param_task_name, task_path)  # need to get taskrun name from parent path
             else:
                 workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
                     task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_param.get('name'))
@@ -599,24 +657,40 @@ def big_data_passing_tasks(prname: str, task: dict, pipelinerun_template: dict,
         if task_spec.get('results', []):
             copy_results_artifact_step = _get_base_step('copy-results-artifacts')
             copy_results_artifact_step['onError'] = 'continue'  # supported by v0.27+ of tekton.
-            copy_results_artifact_step['script'] += 'TOTAL_SIZE=0\n'
+            script = "set -exo pipefail\nTOTAL_SIZE=0\n"
             for result in task_spec['results']:
                 if task['name'] in artifact_items:
                     artifact_i = artifact_items[task['name']]
                     for index, artifact_tuple in enumerate(artifact_i):
                         artifact_name, artifact = artifact_tuple
                         src = artifact
-                        dst = '$(results.%s.path)' % sanitize_k8s_name(result['name'])
+                        dst = '$(results.%s.path)' % sanitize_k8s_name(result['name'], allow_capital=True)
                         if artifact_name == result['name'] and src != dst:
                             add_copy_results_artifacts_step = True
-                            copy_results_artifact_step['script'] += (
-                                    'ARTIFACT_SIZE=`wc -c %s | awk \'{print $1}\'`\n' % src +
-                                    'TOTAL_SIZE=$( expr $TOTAL_SIZE + $ARTIFACT_SIZE)\n' +
+                            total_size_command = 'ARTIFACT_SIZE=`wc -c %s${SUFFIX} | awk \'{print $1}\'`\n' % src + \
+                                                 'TOTAL_SIZE=$( expr $TOTAL_SIZE + $ARTIFACT_SIZE)\n'
+                            copy_command = '    cp ' + src + ' ' + dst + '\n'
+                            if env.get('OUTPUT_PREVIEW', 'false').lower() == 'true':
+                                preview_size = env.get('OUTPUT_PREVIEW_SIZE', '100')
+                                total_size_command = 'TOTAL_SIZE=$( expr $TOTAL_SIZE + %s)\n' % preview_size
+                                copy_command = '    dd if=' + src + ' of=' + \
+                                               dst + ' bs=' + preview_size + ' count=1\n'
+                            script += (
+                                    'if [ -d ' + src + ' ]; then\n' +
+                                    '  tar -czvf ' + src + '.tar.gz ' + src + '\n' +
+                                    '  SUFFIX=".tar.gz"\n' +
+                                    'fi\n' +
+                                    total_size_command +
                                     'touch ' + dst + '\n' +  # create an empty file by default.
                                     'if [[ $TOTAL_SIZE -lt 3072 ]]; then\n' +
-                                    '  cp ' + src + ' ' + dst + '\n' +
+                                    '  if [ -d ' + src + ' ]; then\n' +
+                                    '    tar -tzf ' + src + '.tar.gz > ' + dst + '\n' +
+                                    '  elif ! awk "/[^[:print:]]/{f=1} END{exit !f}" %s; then\n' % src +
+                                    copy_command +
+                                    '  fi\n' +
                                     'fi\n'
                             )
+            copy_results_artifact_step['command'].append(script)
             _append_original_pr_name_env_to_step(copy_results_artifact_step)
             if add_copy_results_artifacts_step:
                 task['taskSpec']['steps'].append(copy_results_artifact_step)
@@ -625,7 +699,8 @@ def big_data_passing_tasks(prname: str, task: dict, pipelinerun_template: dict,
     task.get("taskSpec", {})['params'] = [
         param for param in task_spec.get('params', [])
         if (task_name, param.get('name')) not in inputs_tasks or
-            param.get('name').endswith("-trname")
+            param.get('name').endswith("-trname") or
+            param.get('name').endswith(OUTPUT_RESULT_PATH_SUFFIX)
     ]
 
     # Remove artifacts from task_spec
@@ -640,17 +715,19 @@ def input_artifacts_tasks_pr_params(template: dict, artifact: dict) -> dict:
     task_name = template.get('name')
     task_spec = template.get('taskSpec', {})
     task_params = task_spec.get('params', [])
+    script = "set -exo pipefail\n"
     for task_param in task_params:
         # For pipeline parameter input artifacts, it will never come from another task because pipeline
         # params are global parameters. Thus, task_name will always be the executing task name.
         workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
             task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_param.get('name'))
         if 'raw' in artifact:
-            copy_inputs_step['script'] += 'mkdir -p %s\n' % pathlib.Path(workspaces_parameter).parent
-            copy_inputs_step['script'] += 'echo -n "%s" > %s\n' % (
+            script += 'mkdir -p %s\n' % pathlib.Path(workspaces_parameter).parent
+            script += 'echo -n "%s" > %s\n' % (
                 artifact['raw']['data'], workspaces_parameter)
         _append_original_pr_name_env(template)
 
+    copy_inputs_step['command'].append(script)
     template['taskSpec']['steps'] = _prepend_steps(
         [copy_inputs_step], template['taskSpec']['steps'])
 
@@ -664,21 +741,29 @@ def input_artifacts_tasks(template: dict, artifact: dict) -> dict:
     volume_template = []
     mounted_param_paths = []
     copy_inputs_step = _get_base_step('copy-inputs')
+    has_copy_input = False
+    copy_input_index = 0
     if 'raw' in artifact:
-        copy_inputs_step['script'] += 'echo -n "%s" > %s\n' % (
-            artifact['raw']['data'], artifact['path'])
+        for step_index in range(len(template['taskSpec']['steps'])):
+            if template['taskSpec']['steps'][step_index].get('name') == 'copy-inputs':
+                copy_input_index = step_index
+                has_copy_input = True
+        if has_copy_input:
+            template['taskSpec']['steps'][copy_input_index]['command'][-1] = \
+            template['taskSpec']['steps'][copy_input_index]['command'][-1] + \
+            'echo -n "%s" > %s\n' % (artifact['raw']['data'], artifact['path'])
+        else:
+            copy_inputs_step['command'].append('set -exo pipefail\necho -n "%s" > %s\n' % (
+                artifact['raw']['data'], artifact['path']))
     mount_path = artifact['path'].rsplit("/", 1)[0]
     if mount_path not in mounted_param_paths:
         _add_mount_path(artifact['name'], artifact['path'], mount_path,
                         volume_mount_step_template, volume_template,
                         mounted_param_paths)
-    template['taskSpec']['steps'] = _prepend_steps(
-        [copy_inputs_step], template['taskSpec']['steps'])
-    # _update_volumes(template, volume_mount_step_template, volume_template)
-    if volume_mount_step_template:
-        template['taskSpec']['stepTemplate'] = {}
-        template['taskSpec']['stepTemplate']['volumeMounts'] = volume_mount_step_template
-        template['taskSpec']['volumes'] = volume_template
+    if not has_copy_input:
+        template['taskSpec']['steps'] = _prepend_steps(
+            [copy_inputs_step], template['taskSpec']['steps'])
+    _update_volumes(template['taskSpec'], volume_mount_step_template, volume_template)
     return template
 
 
@@ -734,3 +819,25 @@ def _append_original_pr_name_env(task_template):
     for step in task_template['taskSpec']['steps']:
         if step['name'] == 'main':
             _append_original_pr_name_env_to_step(step)
+
+
+def fix_big_data_passing_using_volume(workflow, pipeline_conf):
+    if workflow['spec'].get('workspaces') and pipeline_conf.data_passing_method._volume is not None:
+        volume_dict = pipeline_conf.data_passing_method._volume.to_dict()
+        volume_dict_pvc = volume_dict.get('persistent_volume_claim')
+        if volume_dict_pvc:
+            for workspace in workflow['spec']['workspaces']:
+                if workspace['name'] == workflow['metadata']['name']:
+                    workspace.pop('volumeClaimTemplate')
+                    temp_vc = {}
+                    for key, value in volume_dict_pvc.items():
+                        if value != None:
+                            if key == 'claim_name':
+                                temp_vc['claimName'] = value
+                            if key == 'read_only':
+                                temp_vc['readOnly'] = value
+                    workspace['persistentVolumeClaim'] = temp_vc
+                    if pipeline_conf.data_passing_method._path_prefix != None:
+                        workspace['subPath'] = pipeline_conf.data_passing_method._path_prefix
+                    break
+    return workflow
