@@ -53,7 +53,7 @@ def _get_base_step(name: str):
     return {
         'image': TEKTON_BASH_STEP_IMAGE,
         'name': name,
-        'script': '#!/bin/sh\nset -exo pipefail\n'
+        'command': ['sh', '-ec']
     }
 
 
@@ -71,11 +71,11 @@ def _get_copy_result_step_template(step_number: int, result_maps: list):
     """
     args = [""]
     for key in result_maps[step_number].keys():
-        sanitize_key = sanitize_k8s_name(key)
+        sanitize_key = sanitize_k8s_name(key, allow_capital=True)
         args[0] += "mv %s%s $(results.%s.path);\n" % (TEKTON_HOME_RESULT_PATH, sanitize_key, sanitize_key)
     if step_number > 0:
         for key in result_maps[step_number - 1].keys():
-            sanitize_key = sanitize_k8s_name(key)
+            sanitize_key = sanitize_k8s_name(key, allow_capital=True)
             args[0] += "mv $(results.%s.path) %s%s;\n" % (sanitize_key, TEKTON_HOME_RESULT_PATH, sanitize_key)
     return {
         "name": "copy-results-%s" % str(step_number),
@@ -89,6 +89,7 @@ def _get_resourceOp_template(op: BaseOp,
                              name: str,
                              tekton_api_version: str,
                              resource_manifest: str,
+                             owner_reference: str,
                              argo_var: bool = False):
     """Tekton task template for running resourceOp
 
@@ -183,7 +184,16 @@ def _get_resourceOp_template(op: BaseOp,
     if op.attribute_outputs.items():
         template['spec']['results'] = []
         for output_item in sorted(list(op.attribute_outputs.items()), key=lambda x: x[0]):
-            template['spec']['results'].append({'name': output_item[0], 'description': output_item[1]})
+            template['spec']['results'].append({'name': output_item[0], 'type': 'string', 'description': output_item[1]})
+
+    if owner_reference == 'true':
+        owner_reference_env = [
+            {'name': 'POD_NAME', 'valueFrom': {'fieldRef': {'fieldPath': "metadata.name"}}},
+            {'name': 'POD_NAMESPACE', 'valueFrom': {'fieldRef': {'fieldPath': "metadata.namespace"}}},
+            {'name': 'POD_UID', 'valueFrom': {'fieldRef': {'fieldPath': "metadata.uid"}}}
+        ]
+        template['spec']['steps'][0]['env'] = template['spec']['steps'][0].get('env', [])
+        template['spec']['steps'][0]['env'].extend(owner_reference_env)
 
     return template
 
@@ -202,16 +212,18 @@ def _add_mount_path(name: str,
     mounted_param_paths.append(mount_path)
 
 
-def _update_volumes(template: Dict[Text, Any],
+def _update_volumes(template_spec: Dict[Text, Any],
                     volume_mount_step_template: List[Dict[Text, Any]],
                     volume_template: List[Dict[Text, Any]]):
     """
     Update the list of volumes and volumeMounts on a given template
     """
     if volume_mount_step_template:
-        template['spec']['stepTemplate'] = {}
-        template['spec']['stepTemplate']['volumeMounts'] = volume_mount_step_template
-        template['spec']['volumes'] = volume_template
+        template_spec.setdefault('stepTemplate', {})
+        template_spec['stepTemplate'].setdefault('volumeMounts', [])
+        template_spec['stepTemplate']['volumeMounts'].extend(volume_mount_step_template)
+        template_spec.setdefault('volumes', [])
+        template_spec['volumes'].extend(volume_template)
 
 
 def _prepend_steps(prep_steps: List[Dict[Text, Any]], original_steps: List[Dict[Text, Any]]):
@@ -262,9 +274,11 @@ def _process_parameters(processed_op: BaseOp,
     if outputs_dict.get('parameters'):
         template['spec']['results'] = []
         copy_results_step = _get_base_step('copy-results')
+        script = "set -exo pipefail\n"
         for name, path in processed_op.file_outputs.items():
             template['spec']['results'].append({
                 'name': name,
+                'type': 'string',
                 'description': path
             })
             # replace all occurrences of the output file path with the Tekton output parameter expression
@@ -274,7 +288,7 @@ def _process_parameters(processed_op: BaseOp,
                     commands = []
                     for c in s['command']:
                         if path in c:
-                            c = c.replace(path, '$(results.%s.path)' % sanitize_k8s_name(name))
+                            c = c.replace(path, '$(results.%s.path)' % sanitize_k8s_name(name, allow_capital=True))
                             need_copy_step = False
                         commands.append(c)
                     s['command'] = commands
@@ -282,16 +296,15 @@ def _process_parameters(processed_op: BaseOp,
                     args = []
                     for a in s['args']:
                         if path in a:
-                            a = a.replace(path, '$(results.%s.path)' % sanitize_k8s_name(name))
+                            a = a.replace(path, '$(results.%s.path)' % sanitize_k8s_name(name, allow_capital=True))
                             need_copy_step = False
                         args.append(a)
                     s['args'] = args
-                if path == '/tekton/results/' + sanitize_k8s_name(name):
+                if path == '/tekton/results/' + sanitize_k8s_name(name, allow_capital=True):
                     need_copy_step = False
             # If file output path cannot be found/replaced, use emptyDir to copy it to the tekton/results path
             if need_copy_step:
-                copy_results_step['script'] = copy_results_step['script'] + 'cp ' + path + ' $(results.%s.path);' \
-                                                % sanitize_k8s_name(name) + '\n'
+                script = script + 'cp ' + path + ' $(results.%s.path);\n' % sanitize_k8s_name(name, allow_capital=True)
                 mount_path = path.rsplit("/", 1)[0]
                 if mount_path not in mounted_param_paths:
                     _add_mount_path(name, path, mount_path, volume_mount_step_template, volume_template, mounted_param_paths)
@@ -299,6 +312,7 @@ def _process_parameters(processed_op: BaseOp,
             parameter_name = sanitize_k8s_name(processed_op.name + '-' + name, allow_capital_underscore=True, max_length=float('Inf'))
             replaced_param_list.append(parameter_name)
             artifact_to_result_mapping[parameter_name] = name
+        copy_results_step['command'].append(script)
         return copy_results_step
     else:
         return {}
@@ -332,7 +346,7 @@ def _process_output_artifacts(outputs_dict: Dict[Text, Any],
             parameter_name = sanitize_k8s_name(artifact['name'], allow_capital_underscore=True, max_length=float('Inf'))
             artifact_name = artifact_to_result_mapping.get(parameter_name, parameter_name)
             if parameter_name in replaced_param_list:
-                artifact_items.append([artifact_name, "$(results.%s.path)" % sanitize_k8s_name(artifact_name)])
+                artifact_items.append([artifact_name, "$(results.%s.path)" % sanitize_k8s_name(artifact_name, allow_capital=True)])
             else:
                 artifact_items.append([artifact_name, artifact['path']])
                 mount_path = artifact['path'].rsplit("/", 1)[0]
@@ -364,10 +378,12 @@ def _process_base_ops(op: BaseOp):
     """
 
     # map param's (unsanitized pattern or serialized str pattern) -> input param var str
-    map_to_tmpl_var = {
-        (param.pattern or str(param)): '$(inputs.params.%s)' % param.full_name  # Tekton change
-        for param in op.inputs
-    }
+    # Pick the shortest param full name if there's two identical params in the DSL
+    map_to_tmpl_var = {}
+    for param in op.inputs:
+        key = (param.pattern or str(param))
+        if not map_to_tmpl_var.get(key) or len('$(inputs.params.%s)' % param.full_name) < len(map_to_tmpl_var.get(key)):
+            map_to_tmpl_var[key] = '$(inputs.params.%s)' % param.full_name
 
     # process all attr with pipelineParams except inputs and outputs parameters
     for key in op.attrs_with_pipelineparams:
@@ -457,7 +473,12 @@ def _op_to_template(op: BaseOp,
             argo_var = True
 
         # task template
-        template = _get_resourceOp_template(op, processed_op.name, tekton_api_version, manifest, argo_var=argo_var)
+        template = _get_resourceOp_template(op=op,
+                                            name=processed_op.name,
+                                            tekton_api_version=tekton_api_version,
+                                            resource_manifest=manifest,
+                                            owner_reference=op.resource.get('setOwnerReference'),
+                                            argo_var=argo_var)
 
     # initContainers
     if processed_op.init_containers:
@@ -502,7 +523,7 @@ def _op_to_template(op: BaseOp,
                                   artifact_items[op.name])
         if mounted_param_paths:
             template['spec']['steps'].append(copy_results_step)
-        _update_volumes(template, volume_mount_step_template, volume_template)
+        _update_volumes(template['spec'], volume_mount_step_template, volume_template)
 
     # metadata
     if processed_op.pod_annotations or processed_op.pod_labels:
@@ -612,7 +633,7 @@ def _op_to_template(op: BaseOp,
                     for key in result_size_map.keys():
                         # Replace main step results that are not in the first bin to the Tekton home path
                         if key not in verified_result_size_map[0].keys():
-                            sanitize_key = sanitize_k8s_name(key)
+                            sanitize_key = sanitize_k8s_name(key, allow_capital=True)
                             for i, a in enumerate(step['args']):
                                 a = a.replace('$(results.%s.path)' % sanitize_key, '%s%s' % (TEKTON_HOME_RESULT_PATH, sanitize_key))
                                 step['args'][i] = a
@@ -627,6 +648,6 @@ def _op_to_template(op: BaseOp,
         # Update actifact item location to the latest stage in order to properly track and store all the artifacts.
         for i, artifact in enumerate(artifact_items[op.name]):
             if artifact[0] not in verified_result_size_map[step_counter].keys():
-                artifact[1] = '%s%s' % (TEKTON_HOME_RESULT_PATH, sanitize_k8s_name(artifact[0]))
+                artifact[1] = '%s%s' % (TEKTON_HOME_RESULT_PATH, sanitize_k8s_name(artifact[0], allow_capital=True))
                 artifact_items[op.name][i] = artifact
     return template
