@@ -15,7 +15,7 @@
 import json, copy, re
 from kfp_tekton.compiler._k8s_helper import sanitize_k8s_name
 
-from kfp_tekton.tekton import LOOP_GROUP_NAME_LENGTH
+from kfp_tekton.tekton import LOOP_GROUP_NAME_LENGTH, AddOnGroup
 
 
 def _process_argo_vars(workflow):
@@ -73,12 +73,18 @@ def _handle_tekton_pipeline_variables(pipeline_run):
         'pipelineRun-uid': '$(context.pipelineRun.uid)'
     }
 
+    add_type = pipeline_run.get('kind') != 'PipelineRun'
     task_list = pipeline_run['spec']['pipelineSpec']['tasks']
     for task in task_list:
         if task.get('taskRef', {}):
             continue
         if 'taskSpec' in task and 'apiVersion' in task['taskSpec']:
+            # recur for embedded pipeline resources
+            if 'spec' in task['taskSpec'] and 'pipelineSpec' in task['taskSpec']['spec']:
+                resource_spec = task['taskSpec']
+                _handle_tekton_pipeline_variables(resource_spec)
             continue
+
         for key, val in pipeline_variables.items():
             task_str = json.dumps(task['taskSpec']['steps'])
             task_str = _process_argo_vars(task_str)
@@ -90,11 +96,17 @@ def _handle_tekton_pipeline_variables(pipeline_run):
                         task['params'].append({'name': key, 'value': val})
                 else:
                     task['params'] = [{'name': key, 'value': val}]
-                if task['taskSpec'].get('params', ''):
-                    if {'name': key} not in task['taskSpec']['params']:
-                        task['taskSpec']['params'].append({'name': key})
-                else:
-                    task['taskSpec']['params'] = [{'name': key}]
+
+                found = False
+                for param in task['taskSpec'].get('params', []):
+                    if param['name'] == key:
+                        found = True
+                        break
+                if not found:
+                    param = {'name': key}
+                    if add_type:
+                        param['type'] = 'string'
+                    task['taskSpec'].setdefault('params', []).append(param)
 
     return pipeline_run
 
@@ -229,6 +241,13 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                 custom_task_cr['spec']['parallelism'] = custom_task[custom_task_key]['spec']['parallelism']
                 # remove from pipeline run spec
                 del custom_task[custom_task_key]['spec']['parallelism']
+            if custom_task[custom_task_key].get('iteration_number') is not None:
+                # enumerate() is used, TektonLoopIterationNumber is injected, get its name and remove it from params
+                custom_task_cr['spec']['iterationNumberParam'] = custom_task[custom_task_key]['iteration_number']
+                custom_task[custom_task_key]['spec']['params'] = [custom_task_param for custom_task_param
+                                                                  in custom_task[custom_task_key]['spec']['params']
+                                                                  if custom_task_param['name'] !=
+                                                                      custom_task_cr['spec']['iterationNumberParam']]
             custom_task_cr['spec']['iterateParam'] = custom_task[custom_task_key]['loop_args']
             separator = custom_task[custom_task_key].get('separator')
             if separator is not None:
@@ -258,6 +277,22 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                     if param['name'] != separator_param
                 ]
 
+        if custom_task[custom_task_key]['kind'] == 'addon':
+            # remove params that are created for internal use. those params have `addon_param` attribute
+            custom_task[custom_task_key]['spec']['params'] = [custom_task_param for custom_task_param
+                in custom_task[custom_task_key]['spec']['params']
+                if not hasattr(custom_task[custom_task_key]['_data'].params.get(custom_task_param['name'], {}), 'addon_param')]
+            # clean up params in spec.taskSpec.spec.pipelineSpec.params, remove those
+            # are not used by downstream Ops/OpsGroups. one-level down only
+            child_task_params = []
+            for task in custom_task_cr['spec']['pipelineSpec']['tasks']:
+                for task_param in task.get('params', []):
+                    if task_param['value'].startswith('$(') and task_param['value'].endswith(')'):
+                        child_task_params.append(task_param['name'])
+            custom_task_cr['spec']['pipelineSpec']['params'] = [value for value
+                                                    in custom_task_cr['spec']['pipelineSpec']['params']
+                                                    if value['name'] in child_task_params]
+
         # need to process task parameters to replace out of scope results
         # because nested graph cannot refer to task results outside of the sub-pipeline.
         custom_task_cr_task_names = [custom_task_cr_task['name'] for custom_task_cr_task in custom_task_cr['spec']['pipelineSpec']['tasks']]
@@ -267,9 +302,14 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                     param_results = re.findall('\$\(tasks.([^ \t\n.:,;\{\}]+).results.([^ \t\n.:,;\{\}]+)\)', task_param['value'])
                     for param_result in param_results:
                         if param_result[0] not in custom_task_cr_task_names:
-                            task['params'] = json.loads(
-                                json.dumps(task['params']).replace(task_param['value'],
-                                '$(params.%s-%s)' % param_result))
+                            if param_result[0] == 'None':
+                                task['params'] = json.loads(
+                                    json.dumps(task['params']).replace(task_param['value'],
+                                    '$(params.%s)' % param_result[1]))
+                            else:
+                                task['params'] = json.loads(
+                                    json.dumps(task['params']).replace(task_param['value'],
+                                    '$(params.%s-%s)' % param_result))
         custom_task_crs.append(custom_task_cr)
         custom_task[custom_task_key]['spec']['params'] = sorted(custom_task[custom_task_key]['spec']['params'],
                                                                           key=lambda k: k['name'])
@@ -280,8 +320,6 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
     nested_custom_tasks = []
     custom_task_crs_namelist = []
     for custom_task_key in custom_task.keys():
-        if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
-            sanitize_k8s_name(custom_task_key, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
         custom_task_crs_namelist.append(custom_task_key)
     for custom_task_key in custom_task.keys():
         for inner_task_name in custom_task[custom_task_key]['task_list']:
@@ -329,6 +367,19 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                             custom_task_cr_again['spec']['pipelineSpec']['params'] = sorted(
                                 custom_task_cr_again['spec']['pipelineSpec']['params'], key=lambda k: k['name'])
                 # add children params to the root tasks
+                global_task_values = set()
+                all_nested_loop = []
+                nested_task = nested_custom_task['nested_custom_task']
+                while nested_task:
+                    all_nested_loop.append(nested_task)
+                    has_nested_task = False
+                    for n_task in nested_custom_tasks:
+                        if n_task['father_ct'] == nested_task:
+                            nested_task = n_task['nested_custom_task']
+                            has_nested_task = True
+                            break
+                    if not has_nested_task:
+                        break
                 for task in tasks:
                     if task['name'] == nested_custom_task['root_ct']:
                         task['params'].extend(copy.deepcopy(nested_custom_task_special_params))
@@ -337,6 +388,27 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                         task['params'].extend(nested_custom_task_special_params)
                     if task.get('params') is not None:
                         task['params'] = sorted(task['params'], key=lambda k: k['name'])
+                        if task['name'] in all_nested_loop:
+                            for param in task['params']:
+                                if '$(params.' in param['value'] and 'subvar-' not in param['value']:
+                                    global_task_values.add(param['value'])
+                # Add any pipeline global params to the nested loop layers
+                all_params = []
+                for custom_param in custom_task_cr['spec']['pipelineSpec']['params']:
+                    all_params.append(''.join(['$(params.', custom_param['name'], ')']))
+                for global_task_value in global_task_values:
+                    if global_task_value not in all_params:
+                        all_params.append(global_task_value)
+                        custom_task_cr['spec']['pipelineSpec']['params'].append(
+                            {'name': re.findall('\$\(params.([^ \t\n.:,;\{\}]+)\)', global_task_value)[0],
+                             'type': 'string'}
+                        )
+                        for task in tasks:
+                            if task['name'] == nested_custom_task['father_ct']:
+                                task['params'].append(
+                                    {'name': re.findall('\$\(params.([^ \t\n.:,;\{\}]+)\)', global_task_value)[0],
+                                     'value': global_task_value}
+                                )
                 for special_param in nested_custom_task_special_params:
                     for nested_param in nested_custom_task_spec['params']:
                         if nested_param['name'] == special_param['name']:
@@ -382,7 +454,16 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                     new_params_spec.append(workflow_param)
                     searched_params.append(workflow_param['name'])
             workflow['spec']['pipelineSpec']['tasks'][i]['params'] = new_params_spec
-    return custom_task_crs, workflow
+    # call post-hook api if there is
+    updated_custom_task_crs = []
+    for cr in custom_task_crs:
+        cr_task = custom_task.get(cr['metadata']['name'])
+        updated_custom_task_crs.append(
+            cr_task.get('_data').post_task_spec(cr)
+            if cr_task and cr_task.get('_data') and
+                isinstance(cr_task.get('_data'), AddOnGroup)
+            else cr)
+    return updated_custom_task_crs, workflow
 
 
 def find_ancestors(nested_custom_tasks: list, father_ct_name, ancestors: list, root_ct):
